static char rcsId[] = "$Header: ggc.c,v 41.1 84/03/14 20:00:48 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 *	All Rights Reserved.
 */

/*
 *	Contains the heart of BS; the memory system.  
 *	The procedures in this file are related to Generation Scavenging.
 */

# include	<stdio.h>
# include	<sys/file.h>

# include	"utils.h"
# include	"mem.h"
# include	"prims.h"
# include	"execute.h"
# include	"cursor.h"

rp_t		FirstFreeContext;	/* head of context free pool */
					/* end of list marked by integer */

/*
 *	The Remembered array contains pointers to old objects that
 *	points to new objects.  These pointers will be used to determine
 *	which object ought to be ggc'ed.
 */

rp_t		Remembered[MAX_REMEMBERED];
rp_t		*NextRemembered;	/* first unused slot in Remembered */


/*
 *	Memory partitions.  See mem.c for details.
 */

struct space	New, Survivor[2], Old;	
u_int		PastSurvivorIndex;


static boolean	ggcCheck(), graphWalk();

static void	ggcRememberedScavenge(), ggcFutureSurvivorsScavenge(),
		ggcRegistryScavenge();
static rp_t	ggcCopy();


/*
 *	Objects known to the interpreter that must be scavenged.
 */

rp_t		DisplayForm, InputSemaphore, SignalAtTickSemaphore,
		LowSpaceSemaphore;


static	boolean	isTimeToGGC;

/*
 *	The next macro is for the scavenge routines.
 *	It has been optimized to save calls on ggcCopy.
 *	oldReg is the dividing line between Old space and 
 *	New/PastSurvivors space.
 *	Since we are only interested in refs to new objects,
 *	a single unsigned comparison screens out smallInts and old oops.
 *	Some invokations need to set a flag if a new oop is found.
 *	doflag is passed as the required statement.
 *	This way, routines that do not need this merely leave it blank.
 */

# define GGCTestSetCopyAndForward(src, dst, doflag)\
	if ((u_int)(src) < (u_int)oldReg) /* new or survivors */ {\
		if (!(flagsRp((rp_t)src) & FORWARDED)) /* not moved */\
			(dst) = (oop_t)ggcCopy((rp_t)src);\
		else\
			(dst) = (oop_t)forwardingRp((rp_t)src);\
		doflag;\
	}\
	else if (TRUE)\
		;\
	else

/*
 *	Generation Scavenging
 */

/*
 *	Called periodically (10 times/sec) to force ggc when necessary
 *	Uses feedback-mediated algorithm:
 *		if would run out of room in two more calls,
 *		do ggc.
 */

pollGenerationSpace()
{
	static u_int	lastRoomLeft = 0;	/* space left after prev call */
	static u_int	maxUsed = 0;		/* max mem used between calls */
	u_int		roomLeft;		/* space left right now */

	if (isTimeToGGC)			/* already know it */
		return;

	roomLeft = wordPointerDifference(New.sp_top , New.sp_next);

	if (roomLeft < lastRoomLeft	/* didn't just finish a ggc */
	    &&   maxUsed  <  lastRoomLeft - roomLeft)
		maxUsed = lastRoomLeft - roomLeft;

	if (roomLeft  <  2 * maxUsed) {
		isTimeToGGC = TRUE;
		forcePoll();
	}
	lastRoomLeft = roomLeft;
}


/*
 *	Called at top level to do ggc if needed.
 */

checkGenerationSpace(acp)
rp_t *acp;
{
	if (isTimeToGGC) {
		traceGGCFor("Reclamation");
		ggc(acp);
		isTimeToGGC = FALSE;
	}
}


/*
 *	Main routine for Generation Scavenging, or Generation
 *	Garbage Collection.  General idea: first move all reachable
 *	new objects from PastSurvivors and New to FutureSurvivors,
 *	leaving forwarding blocks behind.  Then traverse all objects 
 *	that reference them, short-circuiting forwarding blocks.
 *
 *	Pointers in Remembered[] are used to determine what objects
 *	points to objects in New and PastSurvivors spaces.
 *	When FutureSurvivors space is full, objects will be moved 
 *	to Old space (and remembered in Remembered[]).  Therefore,
 *	FutureSurvivors and Remembered have to be repeatedly 
 *	scavenged until no more objects are stored in Old space.
 *
 *	Gets acp so it can update the activeContext, which it moves.
 *	NEW one pass algorithm thanks to LPD.
 *
 *	See ``Generation Scavenging: A High Performance Non-Disruptive
 *	Storage Reclamation Algorithm'' by David Ungar for details.
 */

void
ggc(acp)
register rp_t *acp;
{
	register u_int	oldReg = (u_int)Old.sp_bottom;
	register rp_t	*lastNextRemembered;
	register oop_t	*lastFSNext;
	BE_t	savedCursor[cursorBitmapLength];

	ggcCursor(savedCursor);

	/* 
	 *	FutureSurvivors.sp_next is reset to the bottom of
	 *	FS space so that new objects will be moved to FS
	 *	space starting from the bottom.
	 */

	FutureSurvivors.sp_next = FutureSurvivors.sp_bottom;

	/*
	 *	start things off by scavenging all objects used by
	 *	the virtual machine.
	 */

	GGCTestSetCopyAndForward(SmalltalkSysDictAssocOop,
	    *(oop_t *)&SmalltalkSysDictAssocOop,  );
	GGCTestSetCopyAndForward(InputSemaphore, *(oop_t *)&InputSemaphore,);
	GGCTestSetCopyAndForward(SignalAtTickSemaphore,
	    *(oop_t *)&SignalAtTickSemaphore, );
	GGCTestSetCopyAndForward(LowSpaceSemaphore,
	    *(oop_t *)&LowSpaceSemaphore, );
	GGCTestSetCopyAndForward(DisplayForm, *(oop_t *)&DisplayForm,  );
	GGCTestSetCopyAndForward(*acp, *(oop_t *)acp,  );

	ggcRegistryScavenge();

	/*
	 *	also scavenge any new objects ref'd by old ones
	 */

	ggcRememberedScavenge(Remembered);

	/*
	 *	If ggcFutureSurvivorsScavenge ages an object,
	 *	its contents may not be scanned, since 
	 *	ggcFutureSurvivorsScavenge just scans FutureSurvivors.
	 *	Thus when ggcFutureSurvivorsScavenge promotes an object
	 *	we must scan it with ggcRemembered scavenge to preserve
	 *	the objects it uses.
	 *	When ggcRemembered finds a new object to scavenge, we
	 *	must repeat ggcFutureSurvivorsScavenge to preserve the
	 *	new objects' referents in turn.
	 */

	lastFSNext = FutureSurvivors.sp_bottom;

	for (;;) {
		lastNextRemembered = NextRemembered;
		ggcFutureSurvivorsScavenge((rp_t)lastFSNext);
		if (lastNextRemembered == NextRemembered)
			break;

		lastFSNext = FutureSurvivors.sp_next;
		ggcRememberedScavenge(lastNextRemembered);
		if (lastFSNext == FutureSurvivors.sp_next)
			break;
	}

	traceGGC();
	FirstFreeContext = (rp_t)integerObjectOf(0);	/* it's gone! */
	New.sp_next = New.sp_bottom;
	ToggleSurvivors;
	checkSpaceLimits(acp);
	restoreCursor(savedCursor);
}


/*
 *	Traverse registry, scavenging objects in New 
 *	or Survivor space.
 */

static void
ggcRegistryScavenge()
{
	register u_int	oldReg = (u_int)Old.sp_bottom;
	register rp_t	*rpp, *endp;
	register oop_t	oop;

	for (rpp = &Registry[firstRegXoop],
	    endp = &Registry[lastRegXoop];
	    rpp <= endp; ) {
		oop = (oop_t)*rpp++;
		GGCTestSetCopyAndForward(oop, *(oop_t *)&rpp[-1], );
	}
}


/*
 *	Traverse remembered objects.  Scavenge any new objects that
 *	are referenced by remembered objects.
 *	Eliminate remembered objects that do not refer to new ones.
 *	This is done by checking the foundNew flag to see if the current old
 *	object contained a new pointer.
 */

static void
ggcRememberedScavenge(srcp)
register rp_t *srcp;
{
	register rp_t	rp;
	rp_t	*dstp;

	register boolean foundNew = FALSE;
	register u_int oldReg = (u_int)Old.sp_bottom;


#	define ggcRememberedScavengeOperation(src, dst)\
		GGCTestSetCopyAndForward((src), (dst),  foundNew = TRUE)


	for (dstp = srcp;  srcp < NextRemembered; ) {
		rp = *srcp;
		traverseObject(FALSE, ggcRememberedScavengeOperation);
		if (foundNew) {
			*dstp++ = *srcp++;
			foundNew = FALSE;
		}
		else {
			flagsRp(rp) &= ~REMEMBERED;
			++srcp;
		}
	}
	NextRemembered = dstp;
}



/*
 *	Breadth-first scavenging traversal of new objects.
 *	This is tricky for two reasons:
 *	First, instead of adding an object to be traversed onto
 *	a queue it is copied to the end of FutureSurvivorSpace.
 *	Second, the copying updates sp_next, which lets the loop
 *	continue until the whole space is covered.
 *	The only way to overlook an object is if it gets promoted,
 *	in which case it is copied to Old space and remembered.
 *	But, then the check in ggc() will catch it and cause
 *	another turn around the loop there.
 */

static void
ggcFutureSurvivorsScavenge(rp)
register rp_t rp;
{
	register oop_t	**nextp;
	register u_int	oldReg = (u_int)Old.sp_bottom;

#	define ggcFutureSurvivorsScavengeOperation(src, dst)\
		GGCTestSetCopyAndForward((src), (dst), )
	
	nextp = &FutureSurvivors.sp_next;
	while (rp < (rp_t)*nextp)
		traverseObject(TRUE, ggcFutureSurvivorsScavengeOperation);
}


/*
 *	copy a reachable object, normally to FutureSurvivors (unless
 *	there is not enough space in FutureSurvivors or the object is too old,
 *	in which case it
 *	will be copied to Old space).  Leave a forwarding block.
 */

static rp_t
ggcCopy(rp)
register rp_t rp;
{
	register oop_t	*srcp, *endp, *dstp;
	register struct space	*spacep;
	oop_t	forwardingOop;

#	ifdef CONSERATIVE

		/*
		 *	ggcCopy() is only called for objects that 
		 *	have not been moved.
		 */

		if (flagsRp(rp) & FORWARDED)
			fatal("attempt to ggcCopy FORWARDED object", NullOop);

		if ((rp_t)FutureSurvivors.sp_bottom <= rp
		    &&  rp < (rp_t)FutureSurvivors.sp_top)
			fatal("ggcCopy from FutureSurvivorSpace", NullOop);
#	endif

	endRp(rp, endp, oop_t *);

	/*
	 *	Determine destination space.  Contexts can only be
	 *	copied to FutureSurvivors because they must be young
	 *	to obviate store checking for local variables.
	 */

	if ((char *)endp - (char *)rp  
	    <=  (char *)FutureSurvivors.sp_top - (char *)FutureSurvivors.sp_next
	    &&  (flagsRp(rp) & CONTEXT || ++ageRp(rp) <= maxAge))
		spacep = &FutureSurvivors;
	else if (flagsRp(rp) & CONTEXT)
		fatal("ggcCopy: out of new space for context", (oop_t)rp);
	else if ((char *)endp - (char *)rp  
	    <=  (char *)Old.sp_top - (char *)Old.sp_next)
		spacep = &Old;
	else
		fatal("ggcCopy: out of old space", (oop_t)rp);

	forwardingOop = (oop_t)spacep->sp_next;
	for (srcp = (oop_t *)rp,  dstp = spacep->sp_next;  srcp < endp;)
		*dstp++ = *srcp++;
	spacep->sp_next = dstp;

	flagsRp(rp) |= FORWARDED;
	forwardingRp(rp) = (rp_t)forwardingOop;
	if (spacep == &Old) {
		rp = (rp_t)forwardingOop;
		remember(rp);
	}
	traceGGCCopy(rp, wordPointerDifference(endp, (oop_t *)rp));
	return ((rp_t)forwardingOop);
}


/*
 *	ggcCheck audits ggc().  Untested.
 */

static boolean
ggcCheck()
{
	unmarkObjects(&New);
	unmarkObjects(&FutureSurvivors);
	unmarkObjects(&PastSurvivors);
	unmarkObjects(&Old);

	return (graphWalk(SmalltalkSysDictAssocOop));
}


/*
 *	recursive traversal using marks.  Make sure nothing is in
 *	FutureSurvivors that is reachable.  
 *
 *	It checks if the object ``oop'' has the same mark as ``mark''.
 *	They are considered as having the same mark iff all its 
 *	adjacent nodes (class, objects, etc.) have the same mark.
 */

static boolean
graphWalk(rp)
register rp_t rp;
{
	if (isIntegerObject((oop_t)rp))
		return (TRUE);
	if (flagsRp(rp) & MARKED)
		return (TRUE);
	if (FutureSurvivors.sp_bottom <= (oop_t *)rp
	    &&  (oop_t *)rp < FutureSurvivors.sp_top) {
		fprintf(stderr, "oop 0x%x is in wrong space\r\n", rp);
		fflush(stderr);
		return (FALSE);
	}
	flagsRp(rp) |= MARKED;
	
#	define graphWalkOperation(src, dummyDst)\
		if (!graphWalk((rp_t)(src)))\
			return (FALSE);\
		else if (TRUE)\
			;\
		else

	traverseObject(FALSE, graphWalkOperation);
	return (TRUE);
}


/*
 *	Clear mark bit in space pointed by spacep.
 */

void
unmarkObjects(spacep)
register struct space *spacep;
{
	register rp_t next, p;

	for (p = (rp_t)spacep->sp_bottom,  next = (rp_t)spacep->sp_next;
	   p < next; ) {
		flagsRp(p) &= ~MARKED;
		endRp(p, p, rp_t);	/* for() loop re-initialization */
	}
}
