static char rcsId[] = "$Header: procPrims.c,v 41.1 84/03/14 20:03:50 pkl Exp $";

/*
 *	Copyright (C) 1983 by the Regents of University of California 
 */

/*
 *	Process Primitives.
 *
 *	These routines get the acp because it is only stored into the
 *	active process when doing a snapshot.  Otherwise the ac is
 *	just carried along in the interpreter.  So to switch processes,
 *	the ac must be stored, the processes juggled, and *acp
 *	changed.  It would be simpler but maybe slower to stabilize 
 *	the ac before calling any of the routines that might changed it.  
 *	It would make it harder to tell which ones might, though.
 */

# include	<stdio.h>
# include	<math.h>
# include	<sgtty.h>
# include	<signal.h>
# include	<ctype.h>

# include	<sys/types.h>
# include	<sys/vtimes.h>

# include	"utils.h"
# include	"mem.h"
# include	"io.h"
# include	"prims.h"
# include	"execute.h"

# include	<sys/timeb.h>


static	rp_t	removeFirstLinkOfList();

static void	addLastLink(), newActiveProcess(), resume(), suspendActive();


/*
 *	Semaphore signal
 */

rp_t
SemSigPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	synchronousSignal((rp_t)*rp, &ac);
	return (ac);
}


/*
 *	Shared routine to signal a semaphore and cause
 *	a process switch if new process is of higher priority.
 *	Changes the active context on a switch.
 */

void
synchronousSignal(sema, acp)
register rp_t sema;
rp_t *acp;
{
	rp_t	firstln;

	if (sema == nilOop)
		return;
	firstln = firstLinkLL(sema);

	/*
	 *	If there are no processes waiting, increment the sem's 
	 *	excess signal count.
	 *	If there is a process, take it off the list and awaken it.
	 */

	if (firstln == nilOop)
		incrementSmallInteger(excessSignalsSe(sema));
	else
		resume(removeFirstLinkOfList(sema), acp);
}


/*
 *	Semaphore wait
 *	Active process waits on given semaphore.
 *	If there are excess signals, just take one away,
 *	otherwise queue up active process and suspend it.
 */

rp_t
SemWaitPrm(rp, ac)
oop_t *rp; 
rp_t ac;
{
	register rp_t	sema;
	register oop_t	excessSignals;


	sema = (rp_t)*rp;
	getInt(excessSignalsSe(sema), excessSignals, NullRp);

	if (excessSignals > 0)
		excessSignalsSe(sema) = integerObjectOf(excessSignals - 1);
	else {
		addLastLink(activeProcessSc(schedulerOop()), (rp_t)*rp);
		suspendActive(&ac);
	}
	return (ac);
}


/*
 *	Process suspend
 *	Only works on activeProcess.
 */

rp_t
ProcSuspPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	if ((rp_t)*rp != activeProcessSc(schedulerOop()))
		return (NullOop);
	suspendActive(&ac);
	return (ac);
}


/*
 *	Suspend active process, find another to run.
 *	Scan processLists backward to find highest priority 
 *	process on lists.
 */

static void
suspendActive(acp)
rp_t *acp;
{
	register rp_t	processLists;	/* array of them all */
	register rp_t	*startp, *curp;

	processLists = processListsSc(schedulerOop());
	
	/*
	 *	Starts from the last one in the array and searches
	 *	backward for a non-nil pointer.  Resume the first
	 *	one found
	 */

	for (startp = rpFieldRp(processLists),
	      curp = &startp[lengthRp(processLists) - 1];
	    startp <= curp;  --curp)
		if (firstLinkLL(*curp) != nilOop)
			break;

	/*
	 *	This can only happen if there's only one process in system.
	 */

	if (startp > curp)
		fatal("no process to switch to", NullOop);
	newActiveProcess(removeFirstLinkOfList(*curp), acp);
}


/*
 *	Process resume
 *	Interface to resume subroutine.
 */

rp_t
ProcResPrm(rp, ac)
oop_t *rp;
rp_t ac;
{
	resume((rp_t)*rp, &ac);
	return (ac);
}


/*
 *	make proc runable
 */

static void
resume(proc, acp)
rp_t proc, *acp;
{
	register rp_t	actproc;
	register oop_t	actpri, newpri;

	/*
	 *	Get activeProcess and compare priorities
	 */

	actproc = activeProcessSc(schedulerOop());
	actpri = priorityPr(actproc);
	convertToInteger(actpri);

	newpri = priorityPr(proc);
	convertToInteger(newpri);

	/*
	 *	If new process' priority is greater than the active 
	 *	process' priority, switch to the new one.  
	 *	Otherwise, put the new process on the run queue.
	 */

	if (newpri > actpri) {
		addLastLink(actproc,
		    (rp_t)oopFieldRp(processListsSc(schedulerOop()))[
		      actpri - 1]);
		newActiveProcess(proc, acp);
	}
	else
		addLastLink(proc,
		    (rp_t)oopFieldRp(processListsSc(schedulerOop()))[
		      newpri - 1]);
}


/*
 *	get highest priority process and switch to it
 *	Called by resume and suspend active.
 */

static void
newActiveProcess(newProc, acp)
register rp_t newProc, *acp;
{
	register rp_t	actProcr, schedr;
	register oop_t	newcon;
	
	newcon = (oop_t)suspendedContextPr(newProc);
	if ((rp_t)newcon == nilOop) {
		warning("attempt to process switch to nil context");
		stackTrace(*acp);
		return;
	}

	/*
	 *	Save current process into scheduler.
	 */

	(void)fflush(stdout);
	actProcr = activeProcessSc(schedulerOop());
	suspendedContextPr(actProcr) = *acp;
	ggcStore(actProcr, *acp);

	/*
	 *	Wake up new process by storing new process' context
	 *	into scheduler.
	 */

	schedr = schedulerOop();
	activeProcessSc(schedr) = newProc;
	*acp = (rp_t)newcon;
	assert(FutureSurvivors.sp_bottom > (oop_t *)*acp 
	    ||  (oop_t *)*acp >= FutureSurvivors.sp_top,
	    "ac in wrong space");
	ggcStore(schedr, newProc);
}


/*
 *	Removes first link of linked list.
 *	Items on list MUST be processes because we also nil out 
 *	list field of removee.  
 *	This causes strange bugs if neglected.
 *	The list must not be empty.
 */

static rp_t
removeFirstLinkOfList(procList)
register rp_t procList;
{
	register rp_t	firstLink;

	{
		register rp_t	nextLinkOfFirstLink;
	
		firstLink = firstLinkLL(procList);
	
		/*
		 *	Get second thing on list (may be nil) and nil link 
		 *	field of first thing on list.
		 *	If there was only one thing on list, make list empty.  
		 *	Otherwise, starts list with second thing.
		 */

		nextLinkOfFirstLink = nextLinkLi(firstLink);
		nextLinkLi(firstLink) = nilOop;
		if (lastLinkLL(procList) == firstLink) {
			firstLinkLL(procList) = nilOop;
			lastLinkLL(procList) = nilOop;
		}
		else {
			firstLinkLL(procList) = nextLinkOfFirstLink;
			ggcStore(procList, nextLinkOfFirstLink);
		}
	}

	/*
	 *	nil out process's list pointer
	 */

	myListPr(firstLink) = nilOop;
	return (firstLink);
}


/*
 *	Adds a process to end of a linked list of them.
 *	Must set the process's list pointer field.
 */

static void
addLastLink(aProc, aList)
register rp_t aProc, aList;
{
	/*
	 *	If the list is empty, set the first link to point 
	 *	to the proc.
	 *	Otherwise, set the last proc on the queue to point 
	 *	to the new proc.
	 */

	if (firstLinkLL(aList) == nilOop)
		firstLinkLL(aList) = aProc;
	else {
		register rp_t	lastLink;

		lastLink = lastLinkLL(aList);	/* register is faster */
		nextLinkLi(lastLink) = aProc;
		ggcStore(lastLink, aProc);
	}

	/*
	 *	Set lastLink to point at proc.
	 */

	lastLinkLL(aList) = aProc;
	ggcStore(aList, aProc);

	/*
	 *	Set proc's list field
	 */

	myListPr(aProc) = aList;
	ggcStore(aProc, aList);
}
